--[[ 
****************************************************
* Historian add-on for World of Warcraft.
* 
* Collects various interesting but ultimately useless statistics about your WoW characters.
****************************************************
]]

local L = AceLibrary("AceLocale-2.2"):new("Historian");
local LS = AceLibrary("AceLocale-2.2"):new("Historian-Shared");
local abacus = AceLibrary("Abacus-2.0");
local f = AceLibrary("HistorianFormatting-1.0");
local deformat = AceLibrary("Deformat-2.0");

local DEBUG = false;


-- *************************************
-- * Chat commands
-- *************************************
local HistorianOptions = { 
    type='group',
    args = {
        report = {
			type = 'execute',
			name = L["CMD_REPORT_TITLE"],
			desc = L["CMD_REPORT_DESC"],
			func = "Report",
        },
        firstUse = {
			type = 'execute',
			name = LS["CMD_FIRST_USE_TITLE"],
			desc = LS["CMD_FIRST_USE_DESC"],
			func = "FirstUse",
        },
		showTooltip = {
			type = "toggle",
			name = LS["CMD_SHOWTOOLTIP_TITLE"],
			desc = LS["CMD_SHOWTOOLTIP_DESC"],
			get = "IsShowingTooltip",
			set = "ToggleShowingTooltip",
		},
		countPartyKills = {
			type = "toggle",
			name = LS["CMD_COUNTPARTYKILLS_TITLE"],
			desc = LS["CMD_COUNTPARTYKILLS_DESC"],
			get = "IsCountingPartyKills",
			set = "ToggleCountingPartyKills",
		},
		countInstanceKills = {
			type = "toggle",
			name = LS["CMD_COUNTINSTANCEKILLS_TITLE"],
			desc = LS["CMD_COUNTINSTANCEKILLS_DESC"],
			get = "IsCountingInstanceKills",
			set = "ToggleCountingInstanceKills",
		},
    },
}


-- *************************************
-- * Set up Ace
-- *************************************
Historian = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1");

Historian:RegisterChatCommand(L["SLASH-COMMANDS"], HistorianOptions);

Historian:RegisterDB("HistorianDB", "HistorianCharDB");


-- *************************************
-- * Set up defaults
-- *************************************
Historian:RegisterDefaults("profile", {
	showTooltip = true,
	countPartyKills = true,
	countInstanceKills = true,
} )

Historian:RegisterDefaults("char", {
	Loot = { 
		MostMoneyEver = 0,
		ItemsOfQuality = { 0, 0, 0, 0, 0, 0 },
	},
	Advancement = {
		Levels = { },
		RecipesLearned = 0,
		TotalDeaths = 0,
		QuestsCompleted = 0,
		Tradeskills = { },
	},
	Discoveries = {
		Areas = 0,
	},
	EnvironmentalDamage = {
		Drowning = {
			Times = 0,
			TotalDamage = 0,
		},
		Fatique = {
			Times = 0,
			TotalDamage = 0,
		},
		Falling = {
			Times = 0,
			TotalDamage = 0,
		},
		Fire = {
			Times = 0,
			TotalDamage = 0,
		},
		Lava = {
			Times = 0,
			TotalDamage = 0,
		},
		Slime = {
			Times = 0,
			TotalDamage = 0,
		},
	},
	Follow = { 
		Players = {
		},
		MostFollowed = {
			Count = 0,
			Name = "",
		}
	},
	Kills = {
		Favorite = {
			Count = 0,
			Name = "",
		},
		Mobs = {
		},
		Total = 0,
		TotalKillingBlows = 0,
	},
	Consumption = {
		Food = 0,
		Drink = 0,
		Tipsy = 0,
		Drunk = 0,
		Smashed = 0,
	},
	Spells = {
		Favorite = {
			Count = 0,
			Name = "",
		},
		Casts = {
		},
		Total = 0,
	},
	Jumps = 0,
	FirstUse = nil,
	DataVersion = 0,
} )


-- *************************************
-- * Properties
-- *************************************
function Historian:IsShowingTooltip()
	return self.db.profile.showTooltip;
end
 
function Historian:ToggleShowingTooltip()
	self.db.profile.showTooltip = not self.db.profile.showTooltip;
end


function Historian:IsCountingPartyKills()
	return self.db.profile.countPartyKills;
end
 
function Historian:ToggleCountingPartyKills()
	self.db.profile.countPartyKills = not self.db.profile.countPartyKills;
end


function Historian:IsCountingInstanceKills()
	return self.db.profile.countInstanceKills;
end
 
function Historian:ToggleCountingInstanceKills()
	self.db.profile.countInstanceKills = not self.db.profile.countInstanceKills;
end


-- *************************************
-- * DebugPrint
-- * Prints a chat message, provided the DEBUG variable is set to true
-- *************************************
function Historian:DebugPrint(message)
	if (DEBUG) then
		self:Print("(DEBUG) " .. message);
	end;
end;



-- *************************************
-- * OnInitialize
-- * Fires when the addon is first loaded
-- *************************************
function Historian:OnInitialize()
	self:MigrateData();	

	self.isEating = false;
	self.isDrinking = false;
	
	if (self.db.char.FirstUse == nil) then
		self.db.char.FirstUse = date();
	end;
end;


-- *************************************
-- * MigrateData
-- * Imports data stored in old formats
-- *************************************
function Historian:MigrateData()
	-- Migrate old data
	if (HISTORIAN_DATABASE ~= nil) then
		self:Print(L["MSG_IMPORT_OLD_DATA"]);
		
		-- Loot stats
		self.db.char.Loot.MostMoneyEver = f:AsNumber(HISTORIAN_DATABASE["LootStats"]["MostMoneyEver"]);
		
		for i = 1, 6, 1 do
			self.db.char.Loot.ItemsOfQuality[i] = f:AsNumber(HISTORIAN_DATABASE["LootStats"]["LootedItemsOfQuality" .. tostring(i)]);
		end;
		
		-- Levels
		if (HISTORIAN_DATABASE["AdvancementStats"] ~= nil) then
			for i = 1, 70, 1 do
				local key = "Level_" .. tostring(i);

				if (HISTORIAN_DATABASE["AdvancementStats"][key] ~= nil) then
					self.db.char.Advancement.Levels[i] = HISTORIAN_DATABASE["AdvancementStats"][key];
				end;
			end;
		end;
		
		-- Total deaths
		self.db.char.Advancement.TotalDeaths = f:AsNumber(HISTORIAN_DATABASE["DeathStats"]["Deaths"]);
		
		-- Quests completed
		self.db.char.Advancement.QuestsCompleted= f:AsNumber(HISTORIAN_DATABASE["QuestStats"]["QuestsCompleted"]);
		
		-- Other advancements
		self.db.char.Advancement.RecipesLearned = f:AsNumber(HISTORIAN_DATABASE["AdvancementStats"]["RecipesLearned"]);
		
		-- Discoveries
		self.db.char.Discoveries.Areas = f:AsNumber(HISTORIAN_DATABASE["AdvancementStats"]["AreasDiscovered"]);
		
		-- Environmental damage
		self.db.char.EnvironmentalDamage.Drowning.Times = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageDrowningCount"]);
		self.db.char.EnvironmentalDamage.Drowning.TotalDamage = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageDrowningDamage"]);

		self.db.char.EnvironmentalDamage.Fatique.Times = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageFatiqueCount"]);
		self.db.char.EnvironmentalDamage.Fatique.TotalDamage = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageFatiqueDamage"]);

		self.db.char.EnvironmentalDamage.Falling.Times = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageFallingCount"]);
		self.db.char.EnvironmentalDamage.Falling.TotalDamage = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageFallingDamage"]);
		
		self.db.char.EnvironmentalDamage.Fire.Times = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageFireCount"]);
		self.db.char.EnvironmentalDamage.Fire.TotalDamage = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageFireDamage"]);
		
		self.db.char.EnvironmentalDamage.Lava.Times = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageLavaCount"]);
		self.db.char.EnvironmentalDamage.Lava.TotalDamage = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageLavaDamage"]);
		
		self.db.char.EnvironmentalDamage.Slime.Times = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageSlimeCount"]);
		self.db.char.EnvironmentalDamage.Slime.TotalDamage = f:AsNumber(HISTORIAN_DATABASE["DamageStats"]["EnvironmentDamageSlimeDamage"]);
		
		-- Follow
		self.db.char.Follow.MostFollowed.Count = f:AsNumber(HISTORIAN_DATABASE["MiscStats"]["MostFollowedCount"]);
		self.db.char.Follow.MostFollowed.Name = HISTORIAN_DATABASE["MiscStats"]["MostFollowedPlayer"];

		for key, value in pairs(HISTORIAN_DATABASE["MiscStats"]) do
			if (key == "MostFollowedCount") then
				self.db.char.Follow.MostFollowed.Count = f:AsNumber(value);
			else 
				if (key == "MostFollowedPlayer") then
					self.db.char.Follow.MostFollowed.Name = value;
				else
					local playerName = string.sub(key, 15);
					self.db.char.Follow.Players[playerName] = f:AsNumber(value);
				end;
			end;
		end;

		-- Kills
		for key, value in pairs(HISTORIAN_DATABASE["KillStats"]) do
			if (key == "_FavoritePreyCount") then
				self.db.char.Kills.Favorite.Count = f:AsNumber(value);
			else 
				if (key == "_FavoritePrey") then
					self.db.char.Kills.Favorite.Name = value;
				else
					if (key == "_MobsKilled") then
						self.db.char.Kills.Total = f:AsNumber(value);
					else
						self.db.char.Kills.Mobs[key] = f:AsNumber(value);
					end;
				end;
			end;
		end;

		

		self:Print(L["MSG_IMPORT_COMPLETE"]);
		HISTORIAN_DATABASE = nil;
	end;
	
	if (self.db.char.DataVersion == 0) then
		-- DataVersion was introduced in version 1.0.1.8
		self:Print(L["MSG_IMPORT_OLD_DATA"]);
		
		-- Move all existing kills to "killing blows".
		if (self.db.char.Kills ~= nil) then
			if (self.db.char.Kills.Mobs ~= nil) then
				local newData = {};
				
				for key, value in pairs(self.db.char.Kills.Mobs) do
					newData[key] = {};
					newData[key].KillingBlows = value;
					newData[key].Any = value;
				end;
				
				self.db.char.Kills.Mobs = newData;
			end;
			
			self.db.char.Kills.TotalKillingBlows = self.db.char.Kills.Total;
		end;
		
		self:Print(L["MSG_IMPORT_COMPLETE"]);
	end;
	
	self.db.char.DataVersion = self.version;
end;


-- *************************************
-- * OnEnable
-- * Fires when the addon is enabled
-- *************************************
function Historian:OnEnable()
	-- Register for events
	self:RegisterEvent("AUTOFOLLOW_BEGIN");
	self:RegisterEvent("CHAT_MSG_COMBAT_HOSTILE_DEATH");
	self:RegisterEvent("CHAT_MSG_COMBAT_SELF_HITS");
	self:RegisterEvent("CHAT_MSG_COMBAT_XP_GAIN");
	self:RegisterEvent("CHAT_MSG_LOOT");
	self:RegisterEvent("CHAT_MSG_SYSTEM");
	self:RegisterEvent("LEARNED_SPELL_IN_TAB");
	self:RegisterEvent("PLAYER_DEAD");
	self:RegisterEvent("PLAYER_LEVEL_UP");
	self:RegisterEvent("PLAYER_MONEY");
	self:RegisterEvent("TRADE_SKILL_SHOW");
	self:RegisterEvent("UI_INFO_MESSAGE");
	self:RegisterEvent("UNIT_AURA");
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT");
	
	self:SecureHook("AscendStop");
	
	self:CreateSpellTable();
end;


-- *************************************
-- * CreateSpellTable
-- * Prepares the self.realSpells variable for use in UNIT_SPELLCAST_SUCCEEDED
-- *************************************
function Historian:CreateSpellTable()
	self.realSpells = {};

	local i, s;
	-- Starts at 2 in order to exclude the General tab
	for i = 2, MAX_SKILLLINE_TABS do
		local name, texture, offset, numSpells = GetSpellTabInfo(i);
		if not name then break; end;

		for s = offset + 1, offset + numSpells do
			local spell, rank = GetSpellName(s, BOOKTYPE_SPELL);
			self.realSpells[spell] = true;
		end;
	end;
end;


-- *************************************
-- * Tooltip
-- *************************************
function Historian:UPDATE_MOUSEOVER_UNIT()
	if GameTooltip:IsVisible() and GameTooltipTextLeft1:GetText() ~= nil then
		if (not self.db.profile.showTooltip) then
			return;
		end;

		if (UnitIsPlayer("mouseover")) or (not UnitCreatureType("mouseover")) or (UnitIsFriend("player","mouseover")) then 
			return;
		end;

		local mobName = UnitName("mouseover");

		if (mobName ~= nil) then
			local killStats = self.db.char.Kills.Mobs[mobName];
			if (killStats ~= nil) then
				local numberOfKills = f:AsNumber(killStats.Any);
				local numberOfKBs = f:AsNumber(killStats.KillingBlows);

				if (numberOfKills > 0) then
					local red = 1;
					local green = 1;
					local blue = 0;

					if (numberOfKills == 1) then
						GameTooltip:AddLine(L["MSG_TOOLTIP_KILLED_SINGLE"], red, green, blue);
					else
						if (numberOfKills ~= numberOfKBs) then
							GameTooltip:AddLine(string.format(L["MSG_TOOLTIP_KILLED_WITH_KB"], f:IntToStr(numberOfKills), f:IntToStr(numberOfKBs)), red, green, blue);
						else
							GameTooltip:AddLine(string.format(L["MSG_TOOLTIP_KILLED"], f:IntToStr(numberOfKills)), red, green, blue);
						end;
					end;
					GameTooltip:Show();
				end;
			end;
		end;
	end
end;



-- *************************************
-- * ReportEnvDamage
-- *************************************
function Historian:ReportEnvDamage(type, times, damage)
	local timesText = f:IntToStr(times);
	local damageText = f:IntToStr(damage);
	local formatted = "";
	
	if (times == 1) then
		formatted = string.format(LS["STAT_ENV_VALUE_SINGLE"], damageText);
	else
		formatted = string.format(LS["STAT_ENV_VALUE"], timesText, damageText);
	end;
	
	self:Print(type .. ": " .. formatted);
end;


-- *************************************
-- * ReportMostFollowed
-- *************************************
function Historian:ReportMostFollowed(player, times)
	local formatted = "";

	if (times == 1) then
		formatted = string.format(LS["STAT_MOST_FOLLOWED_VALUE_SINGLE"], player);
	else
		formatted = string.format(LS["STAT_MOST_FOLLOWED_VALUE"], player, f:IntToStr(times));
	end;

	self:Print(LS["STAT_MOST_FOLLOWED"] .. ": " .. formatted);
end;


-- *************************************
-- * Report
-- *************************************
function Historian:Report()
	self:Print(L["MSG_REPORT_HEADER"]);
	self:Print(string.upper(LS["STAT_CATEGORY_FORTUNE"]));
	self:Print(LS["STAT_MAX_MONEY"] .. ": " .. abacus:FormatMoneyFull(self.db.char.Loot.MostMoneyEver, true));
	self:Print(L["MSG_REPORT_SEPARATOR"]);
	
	self:Print(string.upper(LS["STAT_CATEGORY_ITEMS_LOOTED"]));
	self:Print(LS["STAT_LOOTED_COMMON"] .. ": " .. f:IntToStr(self.db.char.Loot.ItemsOfQuality[1]));
	self:Print(LS["STAT_LOOTED_UNCOMMON"] .. ": " .. f:IntToStr(self.db.char.Loot.ItemsOfQuality[2]));
	self:Print(LS["STAT_LOOTED_RARE"] .. ": " .. f:IntToStr(self.db.char.Loot.ItemsOfQuality[3]));
	self:Print(LS["STAT_LOOTED_EPIC"] .. ": " .. f:IntToStr(self.db.char.Loot.ItemsOfQuality[4]));
	
	if (f:AsNumber(self.db.char.Loot.ItemsOfQuality[5]) > 0) or DEBUG then
		self:Print(LS["STAT_LOOTED_LEGENDARY"] .. ": " .. f:IntToStr(self.db.char.Loot.ItemsOfQuality[5]));
	end;

	if (f:AsNumber(self.db.char.Loot.ItemsOfQuality[6]) > 0) or DEBUG then
		self:Print(LS["STAT_LOOTED_ARTIFACT"] .. ": " .. f:IntToStr(self.db.char.Loot.ItemsOfQuality[6]));
	end;

	self:Print(L["MSG_REPORT_SEPARATOR"]);
	
	self:Print(string.upper(LS["STAT_CATEGORY_ACHIEVEMENTS"]));
	self:Print(LS["STAT_QUESTS_COMPLETED"] .. ": " .. f:IntToStr(self.db.char.Advancement.QuestsCompleted));
	self:Print(LS["STAT_RECIPES_LEARNED"] .. ": " .. f:IntToStr(self.db.char.Advancement.RecipesLearned));
	self:Print(LS["STAT_AREAS_DISCOVERED"] .. ": " .. f:IntToStr(self.db.char.Discoveries.Areas));
	self:Print(LS["STAT_TOTAL_KILLS"] .. ": " .. f:IntToStr(self.db.char.Kills.Total));
	self:Print(L["MSG_REPORT_SEPARATOR"]);

	self:Print(string.upper(LS["STAT_CATEGORY_DISAPPOINTMENTS"]));

	if (f:AsNumber(self.db.char.EnvironmentalDamage.Drowning.Times) > 0) or DEBUG then
		self:ReportEnvDamage(LS["STAT_ENV_DROWNING"], self.db.char.EnvironmentalDamage.Drowning.Times, self.db.char.EnvironmentalDamage.Drowning.TotalDamage);
	end;
	if (f:AsNumber(self.db.char.EnvironmentalDamage.Fatique.Times) > 0) or DEBUG then
		self:ReportEnvDamage(LS["STAT_ENV_FATIGUE"], self.db.char.EnvironmentalDamage.Fatique.Times, self.db.char.EnvironmentalDamage.Fatique.TotalDamage);
	end;
	if (f:AsNumber(self.db.char.EnvironmentalDamage.Falling.Times) > 0) or DEBUG then
		self:ReportEnvDamage(LS["STAT_ENV_FALL"], self.db.char.EnvironmentalDamage.Falling.Times, self.db.char.EnvironmentalDamage.Falling.TotalDamage);
	end;
	if (f:AsNumber(self.db.char.EnvironmentalDamage.Fire.Times) > 0) or DEBUG then
		self:ReportEnvDamage(LS["STAT_ENV_FIRE"], self.db.char.EnvironmentalDamage.Fire.Times, self.db.char.EnvironmentalDamage.Fire.TotalDamage);
	end;
	if (f:AsNumber(self.db.char.EnvironmentalDamage.Lava.Times) > 0) or DEBUG then
		self:ReportEnvDamage(LS["STAT_ENV_LAVA"], self.db.char.EnvironmentalDamage.Lava.Times, self.db.char.EnvironmentalDamage.Lava.TotalDamage);
	end;
	if (f:AsNumber(self.db.char.EnvironmentalDamage.Slime.Times) > 0) or DEBUG then
		self:ReportEnvDamage(LS["STAT_ENV_SLIME"], self.db.char.EnvironmentalDamage.Slime.Times, self.db.char.EnvironmentalDamage.Slime.TotalDamage);
	end;

	self:Print(LS["STAT_DEATHS"] .. ": " .. string.format(LS["STAT_DEATHS_VALUE"], f:IntToStr(self.db.char.Advancement.TotalDeaths)));
	
	self:Print(L["MSG_REPORT_SEPARATOR"]);
	
	self:Print(string.upper(LS["STAT_CATEGORY_HABITS"]));
	
	if (f:AsNumber(self.db.char.Follow.MostFollowed.Count) > 0) or DEBUG then
		self:ReportMostFollowed(self.db.char.Follow.MostFollowed.Name, self.db.char.Follow.MostFollowed.Count);
	end;
	
	if (f:AsNumber(self.db.char.Kills.Favorite.Count) > 0) or DEBUG then
		self:Print(LS["STAT_FAVORITE_PREY"] .. ": " .. string.format(LS["STAT_FAVORITE_PREY_VALUE"], self.db.char.Kills.Favorite.Name, f:IntToStr(self.db.char.Kills.Favorite.Count)));
	end;	

	if (f:AsNumber(self.db.char.Spells.Favorite.Count) > 0) or DEBUG then
		self:Print(LS["STAT_FAVORITE_SPELL"] .. ": " .. string.format(LS["STAT_FAVORITE_SPELL_VALUE"], self.db.char.Spells.Favorite.Name, f:IntToStr(self.db.char.Spells.Favorite.Count)));
	end;	

	if (f:AsNumber(self.db.char.Spells.Total) > 0) or DEBUG then
		self:Print(LS["STAT_TOTAL_SPELLS"] .. ": " .. f:IntToStr(self.db.char.Spells.Total));
	end;

	if (f:AsNumber(self.db.char.Consumption.Food) > 0) or DEBUG then
		self:Print(LS["STAT_FOOD"] .. ": " .. string.format(LS["STAT_FOOD_VALUE"], f:IntToStr(self.db.char.Consumption.Food)));
	end;	

	if (f:AsNumber(self.db.char.Consumption.Drink) > 0) or DEBUG then
		self:Print(LS["STAT_DRINK"] .. ": " .. string.format(LS["STAT_DRINK_VALUE"], f:IntToStr(self.db.char.Consumption.Drink)));
	end;	
	
	if (f:AsNumber(self.db.char.Consumption.Tipsy) > 0) or DEBUG then
		self:Print(LS["STAT_TIPSY"] .. ": " .. string.format(LS["STAT_TIPSY_VALUE"], f:IntToStr(self.db.char.Consumption.Tipsy)));
	end;

	if (f:AsNumber(self.db.char.Consumption.Drunk) > 0) or DEBUG then
		self:Print(LS["STAT_DRUNK"] .. ": " .. string.format(LS["STAT_DRUNK_VALUE"], f:IntToStr(self.db.char.Consumption.Drunk)));
	end;
	
	if (f:AsNumber(self.db.char.Consumption.Smashed) > 0) or DEBUG then
		self:Print(LS["STAT_SMASHED"] .. ": " .. string.format(LS["STAT_SMASHED_VALUE"], f:IntToStr(self.db.char.Consumption.Smashed)));
	end;
	

	if (f:AsNumber(self.db.char.Jumps) > 0) or DEBUG then
		self:Print(LS["STAT_JUMPS"] .. ": " .. f:IntToStr(self.db.char.Jumps));
	end;	

	self:Print(L["MSG_REPORT_SEPARATOR"]);
	self:Print(L["MSG_REPORT_NOTE"]);
end;


-- *************************************
-- * FirstUse
-- *************************************
function Historian:FirstUse()
	self:Print(L["MSG_FIRST_USE"], UnitName("player"), self.db.char.FirstUse);
end;


-- *************************************
-- * GetValuesFromChatMessage
-- * Returns the values stored in 'message' based on the Blizzard constant 'localizedMessage'.
-- * For example:
-- * 	local mobKilled = self:GetValuesFromChatMessage(chatMessage, SELFKILLOTHER);
-- * where chatMessage contains 'You have slain SomeEnemy!', returns 'SomeEnemy'.
-- *************************************
function Historian:GetValuesFromChatMessage(message, localizedMessage)
	return deformat:Deformat(message, localizedMessage);
end;


-- *************************************
-- * IsWoWMessage
-- * Returns true if 'message' is of the type 'constant' 
-- * ('constant' contains placeholders so a straight comparison would not work)
-- *************************************
function Historian:IsWoWMessage(message, constant)
	local formatted, replaced, replacedDig;

	formatted, replaced = string.gsub(constant, "%%s", "(.+)");
	formatted, replacedDig = string.gsub(formatted, "%%d", "(%%d%+)");

	return (string.find(message, formatted) == 1);
end;



-- *************************************
-- * PLAYER_MONEY (Remember the biggest fortune the player ever had at any point in time)
-- *************************************
function Historian:PLAYER_MONEY()
	local highestAmount = self.db.char.Loot.MostMoneyEver;
	local currentMoney = GetMoney();

	if (currentMoney > highestAmount) then
		self.db.char.Loot.MostMoneyEver = currentMoney;
		self:TriggerEvent("Historian_Updated");
	end;
end;


-- *************************************
-- * RecordItemLooted (Player received an item)
-- *************************************
function Historian:RecordItemLooted(item)
	if (item == nil) then return; end;
	
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item);

	if (itemRarity > 0) then
		local totalLooted = f:AsNumber(self.db.char.Loot.ItemsOfQuality[itemRarity]) + 1;
		self.db.char.Loot.ItemsOfQuality[itemRarity] = totalLooted;
		self:TriggerEvent("Historian_Updated");
	end;
end;


-- *************************************
-- * CHAT_MSG_LOOT (Player received an item)
-- *************************************
function Historian:CHAT_MSG_LOOT(chatMessage)
	local count = 1;
	local item = nil;
	
	self:DebugPrint("Loot message: " .. chatMessage);
	
	item = self:GetValuesFromChatMessage(chatMessage, LOOT_ITEM_SELF);
	if (item == nil) then
		item, count = self:GetValuesFromChatMessage(chatMessage, LOOT_ITEM_SELF_MULTIPLE);
	end;

	if (item == nil) then
		item = self:GetValuesFromChatMessage(chatMessage, LOOT_ITEM_CREATED_SELF);
	end;
	
	if (item == nil) then
		item, count = self:GetValuesFromChatMessage(chatMessage, LOOT_ITEM_CREATED_SELF_MULTIPLE);
	end;
	
	if (item == nil) then
		item = self:GetValuesFromChatMessage(chatMessage, LOOT_ITEM_PUSHED_SELF);
	end;
	
	if (item == nil) then
		item = self:GetValuesFromChatMessage(chatMessage, LOOT_ITEM_PUSHED_SELF_MULTIPLE);
	end;
	
	if (count == nil) then count = 1; end;

	if (item ~= nil) then
		self:RecordItemLooted(item);
	end;
end;



-- *************************************
-- * CHAT_MSG_COMBAT_HOSTILE_DEATH (Player killed something)
-- *************************************
function Historian:CHAT_MSG_COMBAT_HOSTILE_DEATH(chatMessage)
	local mobKilled = self:GetValuesFromChatMessage(chatMessage, SELFKILLOTHER);
	local killingBlow = (mobKilled ~= nil);
	
	self.ignoreXpGain = false;

	if (killingBlow) then
		self.ignoreUnitDeath = true; -- Don't count the next "(unit) dies." message (which is always added after this one)
		self.ignoreXpGain = true;
		self:RecordKill(mobKilled, killingBlow);
		return;
	end;


	if (self.ignoreUnitDeath) then
		-- This is the second message. The kill has already been recorded due to the player getting the killing blow (see above).
		self.ignoreUnitDeath = false;
		self.ignoreXpGain = true;
		return;
	end;
	
	
	-- Check for party kill
	if (self.db.profile.countPartyKills) then
		local by = nil;
		mobKilled, by = self:GetValuesFromChatMessage(chatMessage, PARTYKILLOTHER);
	end;

	-- Check if the player is in an instance and give credit if that's the case
	if (mobKilled == nil) and (self.db.profile.countInstanceKills) then
		local inInstance, instanceType = IsInInstance();
		if (inInstance) then
			-- Any kill in an instance is counted
			mobKilled = self:GetValuesFromChatMessage(chatMessage, UNITDIESOTHER);
		end;
	end;

	if (mobKilled ~= nil) then
		self:RecordKill(mobKilled, killingBlow);		
		self.ignoreXpGain = true; -- Don't count any "xp gained" messages that might appear directly after this one
	end;
end;


-- *************************************
-- * RecordKill (Player killed something)
-- *************************************
function Historian:RecordKill(mobKilled, killingBlow)
	if (mobKilled ~= nil) then
		if (self.db.char.Kills.Mobs[mobKilled] == nil) then
			self.db.char.Kills.Mobs[mobKilled] = {};
			self.db.char.Kills.Mobs[mobKilled].KillingBlows = 0;
			self.db.char.Kills.Mobs[mobKilled].Any = 0;
		end;
		
		local updatedCount = f:AsNumber(self.db.char.Kills.Mobs[mobKilled].Any) + 1;
		self.db.char.Kills.Mobs[mobKilled].Any = updatedCount;
		
		local favoritePreyKills = self.db.char.Kills.Favorite.Count;
		
		if (updatedCount > favoritePreyKills) then
			self.db.char.Kills.Favorite.Count = updatedCount;

			local favoritePreyName = self.db.char.Kills.Favorite.Name;

			if (favoritePreyName ~= mobKilled) then
				self.db.char.Kills.Favorite.Name = mobKilled;
			end;
		end;
		
		local totalKills = f:AsNumber(self.db.char.Kills.Total) + 1;
		self.db.char.Kills.Total = totalKills;

		if (killingBlow) then
			local updatedKbCount = f:AsNumber(self.db.char.Kills.Mobs[mobKilled].KillingBlows) + 1;
			self.db.char.Kills.Mobs[mobKilled].KillingBlows = updatedKbCount;
			local totalKBs = f:AsNumber(self.db.char.Kills.TotalKillingBlows) + 1;
			self.db.char.Kills.TotalKillingBlows = totalKBs;
		end;
		
		self:TriggerEvent("Historian_Updated");
	end;
end;


-- *************************************
-- * CHAT_MSG_COMBAT_XP_GAIN (Player got experience from killing something)
-- *************************************
function Historian:CHAT_MSG_COMBAT_XP_GAIN(chatMessage)
	if (self.ignoreXpGain) then
		self.ignoreXpGain = false;
		return;
	end;

	local mobKilled = self:GetValuesFromChatMessage(chatMessage, COMBATLOG_XPGAIN_FIRSTPERSON);
	if (mobKilled == nil) then
		mobKilled = self:GetValuesFromChatMessage(chatMessage, COMBATLOG_XPGAIN_EXHAUSTION1);
	end;
	
	if (mobKilled ~= nil) then
		self:RecordKill(mobKilled, false);
	end;
end;


-- *************************************
-- * CHAT_MSG_COMBAT_SELF_HITS (Player inflicted damage on himself)
-- *************************************
function Historian:CHAT_MSG_COMBAT_SELF_HITS(chatMessage)
	if (self:IsWoWMessage(chatMessage, VSENVIRONMENTALDAMAGE_DROWNING_SELF)) then
		return self:OnDrowning(chatMessage);
	end;

	if (self:IsWoWMessage(chatMessage, VSENVIRONMENTALDAMAGE_FATIGUE_SELF)) then
		return self:OnFatique(chatMessage);
	end;

	if (self:IsWoWMessage(chatMessage, VSENVIRONMENTALDAMAGE_FALLING_SELF)) then
		return self:OnFalling(chatMessage);
	end;

	if (self:IsWoWMessage(chatMessage, VSENVIRONMENTALDAMAGE_FIRE_SELF)) then
		return self:OnFireDamage(chatMessage);
	end;

	if (self:IsWoWMessage(chatMessage, VSENVIRONMENTALDAMAGE_LAVA_SELF)) then
		return self:OnLavaSwim(chatMessage);
	end;

	if (self:IsWoWMessage(chatMessage, VSENVIRONMENTALDAMAGE_SLIME_SELF)) then
		return self:OnSlimeSwim(chatMessage);
	end;
end;


-- *************************************
-- * PLAYER_DEAD (Player was killed)
-- *************************************
function Historian:PLAYER_DEAD()
	local totalDeaths = f:AsNumber(self.db.char.Advancement.TotalDeaths) + 1;

	self.db.char.Advancement.TotalDeaths = totalDeaths;

	self:TriggerEvent("Historian_Updated");
end;


-- *************************************
-- * PLAYER_LEVEL_UP (Player gained another level)
-- *************************************
function Historian:PLAYER_LEVEL_UP(newLevel)
	local levelInfo = { };
	levelInfo["Zone"] = GetRealZoneText(); 
	levelInfo["SubZone"] = GetSubZoneText();
	levelInfo["Date"] = date();
	
	self.db.char.Advancement.Levels[newLevel] = levelInfo;
	
	self:TriggerEvent("Historian_Updated");
end;


-- *************************************
-- * CHAT_MSG_SYSTEM (Misc. chat messages)
-- *************************************
function Historian:CHAT_MSG_SYSTEM(chatMessage)
	if (self:IsWoWMessage(chatMessage, ERR_QUEST_COMPLETE_S)) then
		return self:OnQuestCompleted();
	end;

	if (self:IsWoWMessage(chatMessage, ERR_LEARN_RECIPE_S)) then
		return self:OnRecipeLearned();
	end;

	if (self:IsWoWMessage(chatMessage, DRUNK_MESSAGE_ITEM_SELF2)) then
		return self:OnPlayerTipsy();
	end;

	if (self:IsWoWMessage(chatMessage, DRUNK_MESSAGE_ITEM_SELF3)) then
		return self:OnPlayerDrunk();
	end;

	if (self:IsWoWMessage(chatMessage, DRUNK_MESSAGE_ITEM_SELF4)) then
		return self:OnPlayerCompletelySmashed();
	end;

	local count;
	local item = self:GetValuesFromChatMessage(chatMessage, ERR_QUEST_REWARD_ITEM_S);
	
	if (item == nil) then
		count, item = self:GetValuesFromChatMessage(chatMessage, ERR_QUEST_REWARD_ITEM_MULT_IS);
	end;
	
	if (item ~= nil) then
		self:RecordItemLooted(item);
	end;
	
	self:DebugPrint("System message: " .. chatMessage);
end;


-- *************************************
-- * AUTOFOLLOW_BEGIN (Player started following another player)
-- *************************************
function Historian:AUTOFOLLOW_BEGIN(playerFollowed)
	return self:OnFollow(playerFollowed);
end;


-- *************************************
-- * OnFollow (Player started following another player)
-- *************************************
function Historian:OnFollow(playerFollowed)
	local times = f:AsNumber(self.db.char.Follow.Players[playerFollowed]) + 1;
	local currentMax = f:AsNumber(self.db.char.Follow.MostFollowed.Count);
	
	self.db.char.Follow.Players[playerFollowed] = times;

	if (times > currentMax) then
		self.db.char.Follow.MostFollowed.Name = playerFollowed;
		self.db.char.Follow.MostFollowed.Count = times;
		self:TriggerEvent("Historian_Updated");
	end;
end;

	
-- *************************************
-- * OnQuestCompleted (Player completed yet another quest :)
-- *************************************
function Historian:OnQuestCompleted()
	local questsCompleted = f:AsNumber(self.db.char.Advancement.QuestsCompleted) + 1;
	
	self.db.char.Advancement.QuestsCompleted = questsCompleted;
	
	self:TriggerEvent("Historian_Updated");
end;



-- *************************************
-- * OnRecipeLearned (Player learned a new recipe)
-- *************************************
function Historian:OnRecipeLearned()
	local totalKnown = f:AsNumber(self.db.char.Advancement.RecipesLearned) + 1;
	
	self.db.char.Advancement.RecipesLearned = totalKnown;
	
	self:TriggerEvent("Historian_Updated");
end;
	

-- *************************************
-- * OnDiscovery (Player discovered a new zone/area)
-- *************************************
function Historian:OnDiscovery()
	local totalDiscovered = f:AsNumber(self.db.char.Discoveries.Areas) + 1;
	
	self.db.char.Discoveries.Areas = totalDiscovered;
	
	self:TriggerEvent("Historian_Updated");
end;

	
-- *************************************
-- * Environmental damage
-- *************************************
function Historian:RegisterEnvironmentDamage(type, damage)
	local times = f:AsNumber(self.db.char.EnvironmentalDamage[type].Times) + 1;
	local totalDamage = f:AsNumber(self.db.char.EnvironmentalDamage[type].TotalDamage) + f:AsNumber(damage);
	
	self.db.char.EnvironmentalDamage[type].Times = times;
	self.db.char.EnvironmentalDamage[type].TotalDamage = totalDamage;

	self:TriggerEvent("Historian_Updated");
end;
	
	
function Historian:OnDrowning(chatMessage)
	local damage = self:GetValuesFromChatMessage(chatMessage, VSENVIRONMENTALDAMAGE_DROWNING_SELF);
	if (damage ~= nil) then 
		self:RegisterEnvironmentDamage("Drowning", damage);
	end;
end;
	
function Historian:OnFatique(chatMessage)
	local damage = self:GetValuesFromChatMessage(chatMessage, VSENVIRONMENTALDAMAGE_FATIGUE_SELF);
	if (damage ~= nil) then 
		self:RegisterEnvironmentDamage("Fatique", damage);
	end;
end;
	
function Historian:OnFalling(chatMessage)
	local damage = self:GetValuesFromChatMessage(chatMessage, VSENVIRONMENTALDAMAGE_FALLING_SELF);
	if (damage ~= nil) then 
		self:RegisterEnvironmentDamage("Falling", damage);
	end;
end;
	
function Historian:OnFireDamage(chatMessage)
	local damage = self:GetValuesFromChatMessage(chatMessage, VSENVIRONMENTALDAMAGE_FIRE_SELF);
	if (damage ~= nil) then 
		self:RegisterEnvironmentDamage("Fire", damage);
	end;
end;
	
function Historian:OnLavaSwim(chatMessage)
	local damage = self:GetValuesFromChatMessage(chatMessage, VSENVIRONMENTALDAMAGE_LAVA_SELF);
	if (damage ~= nil) then 
		self:RegisterEnvironmentDamage("Lava", damage);
	end;
end;
	
function Historian:OnSlimeSwim(chatMessage)
	local damage = self:GetValuesFromChatMessage(chatMessage, VSENVIRONMENTALDAMAGE_SLIME_SELF);
	if (damage ~= nil) then 
		self:RegisterEnvironmentDamage("Slime", damage);
	end;
end;


-- *************************************
-- * TradeSkillShow
-- *************************************
function Historian:TRADE_SKILL_SHOW()
	if (self.db.char.Advancement.Tradeskills == nil) then
		self.db.char.Advancement.Tradeskills = {};
	end;
	
	local tradeSkillName = GetTradeSkillLine();
	local withHeaders = GetNumTradeSkills();
	
	if (self.db.char.Advancement.Tradeskills[tradeSkillName] == nil) then
		self.db.char.Advancement.Tradeskills[tradeSkillName] = {};
	else
		if (withHeaders == self.db.char.Advancement.Tradeskills[tradeSkillName].WithHeaders) then
			-- No updates required
			return;
		end;
	end;
	
	self.db.char.Advancement.Tradeskills[tradeSkillName].WithHeaders = withHeaders;
	
	-- Remove headers from the equation
	local total = 0;
	local index = 0;
	
	for index = GetFirstTradeSkill(), withHeaders, 1 do
		local skillName, skillType = GetTradeSkillInfo(index);
		if (skillName ~= nil) and (skillType ~= "header") then
			total = total + 1;
		end;
	end;
	
	
	self.db.char.Advancement.Tradeskills[tradeSkillName].Total = total;
	
	self:UpdateRecipeCount();
end;


-- *************************************
-- * UpdateRecipeCount
-- *************************************
function Historian:UpdateRecipeCount()
	local total = 0;
	
	for key, value in pairs(self.db.char.Advancement.Tradeskills) do
		total = total + f:AsNumber(value["Total"]);
	end;

	self.db.char.Advancement.RecipesLearned = total;
	self:TriggerEvent("Historian_Updated");
end;


-- *************************************
-- * UNIT_AURA
-- *************************************
function Historian:UNIT_AURA(unit)
	local wasEating = self.isEating;
	local wasDrinking = self.isDrinking;
	
	self.isEating = false;
	self.isDrinking = false;
	
	if (UnitIsUnit("player", unit)) then
		local i = 1;
		while UnitBuff("player", i) do
			buffName, _, icon, _, duration, timeleft = UnitBuff("player", i);
			
			if (buffName == L["BUFF_FOOD"]) then
				self.isEating = true;
				if (not wasEating) then
					self.db.char.Consumption.Food = self.db.char.Consumption.Food + 1;
					self:TriggerEvent("Historian_Updated");
				end;
			else 
				if (buffName == L["BUFF_DRINK"]) then
					self.isDrinking = true;
					if (not wasDrinking) then
						self.db.char.Consumption.Drink = self.db.char.Consumption.Drink + 1;
						self:TriggerEvent("Historian_Updated");
					end;
				end;
			end;
			
			i = i + 1;
		end
	end;
end;


-- *************************************
-- * AscendStop
-- *************************************
function Historian:AscendStop()
	if (IsFlying() or IsSwimming()) then return; end;
	local now = GetTime();
	
	if (self.lastJump ~= nil) and ((now - self.lastJump) < 0.9) then
		-- A real jump can only occur so often.
		return;
	end;

	self.lastJump = now;	
	self.db.char.Jumps = self.db.char.Jumps + 1;
	self:TriggerEvent("Historian_Updated");
end;


-- *************************************
-- * OnPlayerTipsy
-- *************************************
function Historian:OnPlayerTipsy()
	self.db.char.Consumption.Tipsy = self.db.char.Consumption.Tipsy + 1;
	self:TriggerEvent("Historian_Updated");
end;


-- *************************************
-- * OnPlayerDrunk
-- *************************************
function Historian:OnPlayerDrunk()
	self.db.char.Consumption.Drunk = self.db.char.Consumption.Drunk + 1;
	self:TriggerEvent("Historian_Updated");
end;


-- *************************************
-- * OnPlayerCompletelySmashed
-- *************************************
function Historian:OnPlayerCompletelySmashed()
	self.db.char.Consumption.Smashed = self.db.char.Consumption.Smashed + 1;
	self:TriggerEvent("Historian_Updated");
end;


-- *************************************
-- * UI_INFO_MESSAGE
-- * Called whenever an info message is displayed at the top of the screen
-- *************************************
function Historian:UI_INFO_MESSAGE(message)
	if (self:IsWoWMessage(message, ERR_ZONE_EXPLORED)) then
		return self:OnDiscovery();
	end;
	
	self:DebugPrint(message);	
end;


-- *************************************
-- * LEARNED_SPELL_IN_TAB
-- * Called whenever the player learns a new spell
-- *************************************
function Historian:LEARNED_SPELL_IN_TAB(tabID)
	self:DebugPrint("New spell learned - reloading spells");
	self:CreateSpellTable();
end;


-- *************************************
-- * UNIT_SPELLCAST_SUCCEEDED
-- * Called whenever someone successfully casts a spell nearby (incl. the player)
-- *************************************
function Historian:UNIT_SPELLCAST_SUCCEEDED(unit, spellName, spellRank)
	if (unit ~= "player") then return; end;
	
	Historian:RecordSpellcast(spellName, (self.realSpells[spellName] ~= nil));
end;


-- *************************************
-- * RecordSpellcast (Player cast a spell)
-- *************************************
function Historian:RecordSpellcast(spellName, isRealSpell)
	if (spellName ~= nil) then
		if (self.db.char.Spells.Casts[spellName] == nil) then
			self.db.char.Spells.Casts[spellName] = {};
			self.db.char.Spells.Casts[spellName].Succeeded = 0;
		end;
		
		local updatedCount = f:AsNumber(self.db.char.Spells.Casts[spellName].Succeeded) + 1;
		self.db.char.Spells.Casts[spellName].Succeeded = updatedCount;
		self:DebugPrint("You've cast " .. spellName .. " a total of " .. f:IntToStr(updatedCount) .. " time(s).");

		if not isRealSpell then
			-- This is not really a spell, so it cannot be a favorite
			self:DebugPrint("This is not a usable spell, so it will not count towards the total and it cannot be a favorite.");
			return;
		end;
		
		local favoriteCount = self.db.char.Spells.Favorite.Count;
		
		if (updatedCount > favoriteCount) then
			self.db.char.Spells.Favorite.Count = updatedCount;
			self.db.char.Spells.Favorite.Name = spellName;
			self:DebugPrint(spellName .. " is now your favorite spell.");
		end;
		
		local updatedTotal = f:AsNumber(self.db.char.Spells.Total) + 1;
		self.db.char.Spells.Total = updatedTotal;
		self:DebugPrint("You have cast a total of " .. f:IntToStr(updatedTotal) .. " spell(s).");
		self:TriggerEvent("Historian_Updated");
	end;
end;